# 集群信息收集及故障检测工具

## 介绍

此工具用于集群信息一站式收集、解析和故障诊断，输出告警以及对应解决方案，简化局点、测试问题定位。

### 设计原则

- 通用性：理论上经简单修改后可用于任何集群交付件。
- 易于扩展：模块化设计，添加新的信息收集和检测项无需修改其他代码。
- 易于开发：对http接口调用、命令行指令执行、文件读取、文本解析等操作做了友好封装，代码简洁，可读性高。
- 易于测试：支持通过配置文件进行模拟数据构造，以便于在Windows开发环境上模拟各种生产环境产生的结果。
- 易于调试：自动记录检测过程中的命令执行、http接口调用、文件读取的结果，便于回溯。这些信息也能够作为模拟数据在开发环境调试。

## 开发环境

- Python版本 3.10.6
  - Python新版本有很多好用的特性，请不要使用低于3.10版本的python
- IDE：PyCharm / VSCode
  - 强烈建议使用PyCharm，智能提示和重构功能要比VSCode好用很多。

## 支持平台

由于项目基于Python，Linux版本需要在特定操作系统下编译，因此并不能完全跨平台， 以下环境已经过测试可用
其他版本待后续测试。

- Windows10（仅调试）
- CentOS 6.5 x86_64
- CentOs 7.3 x86_64

## 使用方法

### 开发环境/Windows 中调试

#### 命令行

安装好python后，进入项目路径，命令行下执行如下命令即可

```python
python3 main.py --mock
```

- `--mock` 表示使用模拟数据执行，参见[模拟数据章节](#mock)

#### PyCharm

`菜单` -> `run` -> `Edit Configuration` -> `Parameters`中
添加`--mock` 参数

然后直接右键执行`main.py`即可

### 生产环境/Linux 中运行

将整个项目打包上传到大数据主机任意目录，进入项目根路径执行：

```shell
start.sh
```

此脚本只是对`main.py`的简单包装，同样支持`--mock`参数，如只想收集当前节点的信息，可以添加`--local`参数

> **注意**：
> 
> 为了减小体积，Linux环境中使用的是**精简版Python**，删除了部分目前用不到的标准库文件，如果你引入的标准库在生产环境执行时报找不到，直接去原始Python安装包的`lib`目录下把这个库拷贝过来即可，原始安装包归档在这里。
> 
> \\\nt03\平台开发部\02.部门共享文件夹\99.项目组临时路径\大数据技术部\常用软件归档\14.rpm包\python编译
> 使用python3.10.6-centos7.tgz即可

## 发布包制作方法

将项目拷贝到linux环境

删除临时缓存文件

```python
rm -rf `find . -name '__pycache__'`
```

打包，打包过程中排除模拟数据目录、临时目录、诊断结果、.idea和.svn目录。

```python
tar zcvf cluster-checker.tgz --exclude=mock --exclude=tmp --exclude=output --exclude=venv --exclude=test --exclude=.svn --exclude=.idea .
```


## 程序运行流程

1. 在集群主机上执行主程序
2. 主程序读取大数据配置文件，获取集群所有节点
3. 主程序并行分发自己到所有其他节点
4. 并行启动其他节点的信息收集和诊断
5. 主程序执行本机信息收集和诊断
6. 并行拷贝其他节点的信息收集和诊断结果回主机

## 目录结构

```bash
cluster-checker ............. 项目根路径
  ├─checker ................. 具体信息收集和校验模块目录
  ├─core .................... 核心模块目录（一般不用改）
  ├─mock .................... 模拟数据配置目录，用于生成模拟数据，便于开发环境调试
  ├─test .................... 单元测试用例目录
  │
  │ config.py ............... 项目配置文件
  │ message.py .............. 诊断信息配置文件
  │ main .................... 启动文件模块
  │ start.sh ................ 生产环境启动文件
  │ README.md ............... 项目说明文件
```

## 开发指导

### Checker类

Checker类用于定义集群各种指标信息的收集和诊断。

Checker类必须被定义在`checkers`包中，并继承自`core.checker.BaseChecker`类。程序启动后会自动加载并执行。

一个模块文件中可以定义多个Checker类，为了结构清晰，务必保证**Checker的功能及其所在的模块含义相符**
，同时避免一个Checker实现太多不相关的功能。

### 类结构

以下是一个收集CPU核数信息，并对进行校验输出警告的样例：

```python
class CPU(BaseChecker):

    def __init__(self):
        super().__init__("CPU规格")

    @share
    def cores(self):
        out = self.execute("grep -c 'processor' /proc/cpuinfo")
        return int(out.stdout.strip())

    def run(self) -> None:
        cores = self.cores()
        if cores < config.cpu_min_vcore:
            self.alarms.add(message.核数过少.format(cores, config.cpu_min_vcore))
```

Checker类需要包含有以下要素:

#### 父类

继承`core.checker.BaseChecker`类。 调用父类构造方法：`super().__init__("CPU规格")`入参为校验类的中文描述。

#### 故障诊断方法

实现故障诊断方法：`def run(self)`，该方法用于执行收集节点相关状态并校验健康状态。
如检测到异常，使用`self.alarms.add()`添加告警信息。告警信息会在运行结束后输出在`output目录中`

#### 信息共享

在方法上添加`@share`装饰器可以声明该函数结果可被其他Checker共享。

适用于场景：

1. 某些Checker做故障诊断时需要配合其他Checker收集的信息。
2. 希望某些方法收集到的信息输出到文件中。被`@share`装饰的方法，其返回值会被记录到`output/信息收集`目录中，便于查看。

在Checker中获取其他Checker收集到的信息的方法为：

```python
# CPU为Checker类名，'cores'为需要获取的share方法名
CheckerManager.share(CPU, 'cores')
```

> 需要注意的是，`@share`装饰的方法必须被调用执行过，其结果才会产生，框架并不会自动去调用这些方法。

### 控制Checker在哪些节点运行(暂时没有实现)

Checker默认会在集群所有节点上执行，考虑到某些校验仅支持在集群某些节点上执行。
这种情况下需要实现`should_run_here`方法，用于判断校验类能否在当前节点执行。

```python
def should_run(self) -> bool
```

例如

#### 只在主机执行

我们希望Checker只在主机执行，可以采用如下实现：

```python
def should_run(self) -> bool:
    cluster = CheckerManager.share(Cluster)
    local_ip = cluster.get('db_bigdata.local_ip')
    master_ip = cluster.get('db_bigdata.master')
    return local_ip == master_ip
```

#### 只在最后一台从机上执行

我们希望Checker只在最后一台从机上执行，可以采用如下实现：

```python
def should_run(self) -> bool:
    cluster = CheckerManager.share(Cluster)
    local_ip = cluster.get('db_bigdata.local_ip')
    slaves = cluster.get('db_bigdata.slaves')
    return local_ip == slaves[-1]
```

### 告警文本定义

为了方便复用以及后期维护和修改，告警文本统一维护在`message.py`文件中。
**禁止在具体Checker类中直接创建告警文本对象**

## Checker中常用工具函数

以下工具函数在Checker中很有用

命令行指令

```python
out = self.cmd(f'ping {ip} -c 100 -i 0 -s 10000 -w 10')
```

文件读取

```python
out = self.read_file('/bigdata/salut/service.json')
```

http请求

```python
out = self.http('GET', '202.91.5.102', 9200, '/_cat/tasks')
```

pgsql语句

```python
out = self.pgsql('bigdata', passwd, 'select * from tbl_general_data')
```

greenplum语句

```python
out = self.gpsql('postgres', passwd,
                 "select *, ((now()-query_start)/'1 minute')::integer AS elapsed_hours from pg_stat_activity where state='active'")
```

## 模板化文本解析

信息收集过程会通过命令行指令、http请求、文件读取等方式获取信息，正则提取实现文本解析门槛高，可读性较差。

本项目中，推荐使用TextParser类解析复杂文本。此类是对parse库的包装，但表达式语法没有变，很好理解，参见[https://pypi.org/project/parse/]()

### 用法1

`parse_table_rows`函数适用于表格状文本。入参为行模板，模板匹配文本中所有的行，返回值为所有命中行的列表。

```python
out, err, code = self.execute("fdisk -l")
parser = TextParser(out)
disks = parser.parse_rows('Disk {disk}: {size:f} {unit}, {bytes:d} bytes, {:d} sectors')
```

以上用例结果为

```python
disks = [
    {'disk' = '/dev/sda', 'size' = 10.2, 'unint' = 'GB', 'bytes' = 1000000000},
    {'disk' = '/dev/sdb', 'size' = 20.5, 'unint' = 'GB', 'bytes' = 2000000000}
]
```

### 用法2

`parse`函数适用于不规则的文本。入参可传入多个模板，每个模板都会尝试在文本中进行匹配，但只能命中一次，返回值为所有命中键值对构成的字典。

```python
out, err, code = self.execute("cat /bigdata/salut/conf/db_bigdata.conf")
parser = TextParser(out)
props = parser.parse(
    'roles={roles}',
    'local_ip={local_ip}',
    'master={master}',
    'slave={slaves:comma_split}')
```

以上用例结果为:

```python
props = {
    'roles': 'master',
    'local_ip': '192.168.1.1',
    'master': '192.168.1.1',
    'slaves': ['192.168.1.1', '192.168.1.2']
}
```

## mock功能（模拟数据）

<span id="mock"></span>

mock功能可以模拟生产集群中各种命令的输出。
只需将预期的输出添加到配置文件。 即可在开发环境中快速验证信息收集、解析、诊断功能是否正常。同时也方便IDE中做断点调试。

> 生产环境运行时也会产生信息收集文件，该文件与模拟数据配置文件格式一致，
因此，**你可以将生产环境中收集到的信息拷贝到本地作为模拟数据调试**。

支持：

1. 模拟命令行执行返回的标准输入、标准错误输入和退出码
2. 模拟http接口调用返回的信息和退出码
3. 模拟文件读取的内容

### mock开关

- 修改 `config.py` 文件中 `mock_enable` 配置项为 `True` 即可启用mock
- 也可以在执行程序时添加 `--mock` 参数

### 默认支持mock的函数

BaseChecker中已经内置如下可mock函数，调用这些函数会自动去mock用例中寻找对应的模拟数据。

- 命令行执行：`def cmd(self, cmd)`
- gp语句执行：`def gpsql(self, database, passwd, sql: str, user='gpadmin', port=6543)`
- pgsql语句执行：`def pgsql(self, database, passwd, sql: str, user='postgres', port=5432)`
- ssh远程命令执行：`def ssh(self, ip, cmd)`
- 文件读取：`def read_file(self, path)`
- http请求执行：`http(self, method, ip, port, path, payload='', headers=None)`

### 配置方式

模拟数据配置文件维护在`mock`目录下，
具体路径由`config.py`配置文件中`mock_dir`配置项决定。

以下为模拟`cat /etc/redhat-release`指令执行的结果

```xml
<?xml version="1.0" encoding="utf-8" ?>
<Cases>
    <Case id="cat /etc/redhat-release">
        <Output>
            CentOS release 6.5 (Final)
        </Output>
        <Error></Error>
        <Code>0</Code>
    </Case>
</Cases>
```

- `<Output>`为标准输出
- `<Error>`为标准错误输出，为空时可以省略
- `<Code>`为退出码，为0时可以省略

一个配置文件中可以有多个用例，如果有重复的用例，以第一个找到的为准。

## 附录

## linux python 精简方法

在联网环境编译并打包

```shell
./configure --prefix=/usr/local/python3 --enable-shared  --with-openssl=/usr/local/openssl/ --with-openssl-rpath=auto
make
make install
cd /usr/local/
tar zcvf python3.tgz python3
```

删除无用的文件

```python
/ usr / local / python3
├─bin.................可执行文件路径，只需保留
├─include.............存放头文件，运行时无用，可以删除
├─lib.................python标准库目录
└─share...............用户手册，可以删除

```