import json
from threading import Thread
import traceback
import config
import message
from checkers.cluster import Cluster
from core.checker import BaseChecker, CheckerManager, share
from core.common import log
from core.mock import mock_cmd
from core.textparser import TextParser


class DBServer(BaseChecker):

    def __init__(self):
        super().__init__('大数据服务')
        self.props = None
        self.conf = None
        self.initial_info_collection_file()

    def run(self) -> None:
        self.version()
        self.props = CheckerManager.share(Cluster, 'db_bigdata')
        context = self.read_file(config.service_path)
        self.conf = json.loads(context.content)
        service_state = self.service_state()
        down = list([service for service, running in service_state.items() if not running])
        if down:
            self.alarms.add(message.服务未启动.format(",".join(down)))

    @share
    def service_state(self):
        """
        获取集群服务状态
        :params
        :return: 返回集群状态字典
        """
        return self.get_status('all')

    @share
    def version(self) -> dict:
        """
        获取大数据版本信息
        只在主节点上执行
        :return: 返回版本信息
        """
        out = self.cmd("dbcfgtool.sh -v")
        parser = TextParser(out.stdout)
        version_info = parser.parse(
            'Interior version : {inter_version}',
            'Exterior version : {exter_version}',
            'BUILDTIME        : {buid_time}')
        return version_info


    # 根据名称获取服务配置
    def get_service(self, name):
        for s in self.conf["serviceList"]:
            if s["name"].lower() == name.lower() or (
                    "shortName" in s.keys() and s["shortName"].lower() == name.lower()):
                return s
        raise BaseException("Service or group '" + name + "' is not defined")

    # 如果里面有group name，则将其解析成服务名
    def parse_services(self, name_list):
        rs = set(name_list)
        for name in name_list:
            for s in self.conf["groupList"]:
                if s["name"].lower() == name.lower():
                    rs.remove(s["name"])
                    rs = rs | self.parse_services(set(s["service"]))
        return rs

    def get_check_cmd(self, service):
        return service["pid"] + "| xargs"

    # 状态检查线程
    def status(self, service, service_state):
        name = service["name"]
        check_cmd = self.get_check_cmd(service)
        try:
            out = self.find_pid(check_cmd)
            if out.stdout != None and out.stdout.strip():
                service_state[name] = True
            else:
                service_state[name] = False
        except Exception as e:
            log.error(traceback.format_exc(e))

    @mock_cmd('pidcheck')
    def find_pid(self, check_cmd):
        return self.cmd(check_cmd + " 2>&1")

    # 当前节点是否为服务绑定的角色
    def is_role_match(self, service):
        if self.props['is_ha']:
            role_bind = service["bindHA"]
        else:
            role_bind = service["bind"]
        if role_bind == ["all"]:
            return True
        else:
            intersection = set(role_bind) & set(self.get_roles())
            log.debug("Intersection of " + service["name"] + ": " + str(intersection))
            return intersection

    def get_roles(self):
        return [role.strip() for role in self.props['roles'].split(',')]

    def get_status(self, *names) -> dict:
        service_state = {}
        service_names = self.parse_services(names)
        service_list = [self.get_service(name) for name in service_names]
        service_list = [service for service in service_list if "enable" in service.keys() and service["enable"]]
        threads = []
        for service in service_list:
            try:
                if self.is_role_match(service):
                    name = service["name"]
                    thread = Thread(target=self.status, args=(service, service_state), name="status-" + name)
                    threads.append(thread)
            except Exception as e:
                log.error(traceback.format_exc(e))

        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        return service_state


if __name__ == '__main__':
    dbserver: DBServer = DBServer()
    dbserver.run()
