import config
import message
from core.checker import BaseChecker, CheckerManager, share
from core.textparser import TextParser


class Disk(BaseChecker):

    def __init__(self):
        super().__init__('磁盘信息')
        self.initial_info_collection_file()

    @share
    def disks(self):
        out = self.cmd("fdisk -l | grep -E '^Disk /'")
        parser = TextParser(out.stdout)
        disks = parser.parse_rows(
            'Disk {disk}: {size:f} {unit}, {bytes:d} bytes', full_match=False)

        return disks

    @share
    def mounts(self):
        mounts = {}
        for dir in config.key_dirs.keys():
            out = self.cmd('df -T ' + dir)
            if not out.success:
                continue
            # /dev/sda3 ext4 1031987352 582929876 396612292  60% /
            parser = TextParser(out.stdout,
                                strip_header='Filesystem{:s}Type{:s}1K-blocks{:s}Used{:s}Available{:s}Use%{:s}Mounted{:s}on')
            dev = parser.parse(
                '{dev:S}{:s}{type:S}{:s}{size:d}{:s}{used:d}{:s}{avail:d}{:s}{use_percent:d}%{:s}{mount_on:S}')
            mounts.setdefault(dev['dev'], dev).setdefault('dirs', []).append(dir)

        return mounts

    def run(self) -> None:
        devs = self.mounts()
        disks = self.disks()
        for dev in devs.values():
            dirs_in_dev = dev['dirs']
            expect_size = sum([config.key_dirs[dir]['expect_size_gb'] for dir in dirs_in_dev])
            actual_size = dev['size'] / 1024 / 1024
            dev_name = dev['dev']
            if expect_size > actual_size:
                dir_expect = '\n'.join([f"{dir}: {config.key_dirs[dir]['expect_size_gb']} GB" for dir in dirs_in_dev])
                self.alarms.add(
                    message.磁盘总容量不足.format(dev_name, round(actual_size, 2), expect_size, dir_expect))

            avail = dev['avail'] / 1024 / 1024
            if avail <= config.disk_min_available_gb:
                self.alarms.add(message.磁盘剩余容量不足.format(dev_name, avail, config.disk_min_available_gb))

            use_percent = dev['use_percent']
            if use_percent >= config.disk_max_used_percent:
                self.alarms.add(
                    message.磁盘使用比例过高.format(dev_name, use_percent, config.disk_max_used_percent))
            elif use_percent >= config.disk_warn_used_percent and use_percent < config.disk_max_used_percent:
                self.alarms.add(
                    message.磁盘使用比例较高.format(dev_name, use_percent, config.disk_max_used_percent))


class Smart(BaseChecker):

    def __init__(self):
        super().__init__('SMART信息')
        self.initial_info_collection_file()

    @share
    def smart(self, disks) -> dict:
        smart = {}
        for disk in disks:
            if config.exclude_mapper in disk['disk']:
                continue
            out = self.cmd('smartctl -x ' + disk['disk'])
            parser = TextParser(out.stdout)
            info = parser.parse("SMART support is:{:s}{is_support_smart}")
            smart[disk['disk']] = {'state': ''}
            if info['is_support_smart'] == "Enabled":
                state_v1 = parser.parse('SMART overall-health self-assessment test result: {state}')
                state_v2 = parser.parse('SMART Health Status: {state}')
                if state_v1:
                    smart[disk['disk']] = state_v1
                else:
                    smart[disk['disk']] = state_v2
            elif 'Unavailable' in info['is_support_smart']:
                smart[disk['disk']]['state'] = disk['disk'] + ' - ' + info['is_support_smart']

        return smart

    def run(self) -> None:
        disks: list = CheckerManager.share(Disk, 'disks')
        smart = self.smart(disks)
        for disk, info in smart.items():
            if 'OK' in info['state']:
                continue
            elif 'Unavailable' in info['state']:
                self.alarms.add(message.磁盘不支持SMART命令.format(disk))
            else:
                self.alarms.add(message.磁盘故障.format(disk))


class DiskRead(BaseChecker):

    def __init__(self):
        super().__init__('磁盘写入性能')
        self.initial_info_collection_file()

    @share
    def paths_speed(self):
        path_speed = {}
        dir_to_write = [dirname for dirname, prop in config.key_dirs.items() if
                        'min_write_mb_per_second' in prop.keys()]
        for path in dir_to_write:
            out = self.cmd(f'dd if=/dev/zero of={path}/write-test.dbf bs=4k count=100000 oflag=direct')
            # 注意dd命令的输出不在标准输出，而在标准错误输出
            parser = TextParser(out.stderr)
            path_speed[path] = parser.parse(
                '{size:d} bytes ({} MB) copied, {seconds:f} s, {speed:f} MB/s')
        return path_speed

    def run(self) -> None:
        speeds = self.paths_speed()

        for path, result in speeds.items():
            expect_speed = config.key_dirs[path]['min_write_mb_per_second']
            actual_speed = int(result['speed'])
            if actual_speed < expect_speed:
                self.alarms.add(message.磁盘写入慢.format(path, actual_speed, expect_speed))


class FileSystem(BaseChecker):

    def __init__(self):
        super().__init__('磁盘文件系统')
        self.initial_info_collection_file()

    @share
    def file_system(self):
        out = self.cmd(f'cat /etc/mtab | grep -E ^/dev')
        parser = TextParser(out.stdout)
        result = parser.parse_rows("{partition} {mount_point} {file_system_format} {file_system_status},{options} {dump} {pass}")

        return result

    def run(self) -> None:
        self.file_system()
