import message
from checkers.cluster import Cluster
from core.checker import BaseChecker, CheckerManager, share
from core.textparser import TextParser
import json


class ES(BaseChecker):

    def __init__(self):
        super().__init__('ElasticSearch')
        self.props = None
        self.read_only_indices = []

    @share
    def tasks(self) -> dict:
        """
        只在主节点上执行该方法
        """
        if self.should_run():
            self.initial_info_collection_file()
            out = self.http('GET', self.props['local_ip'], 9200, '/_cat/tasks')
            parser = TextParser(out.output)
            tasks = parser.parse_rows(
                '{action}{:s}{task_id}{:s}{parent_task_id}{:s}{type}{:s}{start_time}{:s}{timestamp}{:s}{running_time}{:s}{ip}{:s}{node}')
            return tasks

        return {}

    @share
    def shards(self) -> dict:
        """
        只在主节点上执行该方法
        """
        if self.should_run():
            self.initial_info_collection_file()
            out = self.http('GET', self.props['local_ip'], 9200, '/_cat/shards?bytes=mb')
            parser = TextParser(out.output)
            shards = parser.parse_rows(
                '{index}{:s}{shard:d}{:s}{type}{:s}{state}{:s}{docs:d}{:s}{store_in_mb:d}{:s}{ip}{:s}{node}')
            return shards

        return {}

    @share
    def indices(self):
        """
        只在主节点上执行该方法
        """
        if self.should_run():
            self.initial_info_collection_file()
            out = self.http('GET', self.props['local_ip'], 9200, '/_cat/indices?bytes=mb')
            parser = TextParser(out.output)
            indices = parser.parse_rows(
                '{health}{:s}{status}{:s}{index}{:s}{uuid}{:s}{shard.num:d}{:s}{replica.num:d}{:s}{docs.count:d}{:s}{docs.deleted:d}{:s}{store.size.mb:d}{:s}{pri.store.size.mb:d}')
            return indices

        return {}

    def get_local_ip(self):
        self.props = CheckerManager.share(Cluster, "db_bigdata")

    def is_read_only(self, index) -> str:
        """
        判断索引是否为只读
        只在主节点上执行该方法

        @param: index 索引
        @return 是否为只读索引
        """
        if self.should_run():
            self.initial_info_collection_file()
            out = self.http('GET', self.props['local_ip'], 9200, '/' + index + '/_settings')
            get_index_json_obj = json.loads(out.output)
            blocks_result = get_index_json_obj[index]["settings"]["index"].get("blocks")
            if blocks_result and (
                    blocks_result.get("read_only") == "true" or blocks_result.get("read_only_allow_delete") == "true"):
                return index

    def run(self) -> None:
        self.get_local_ip()
        indices = self.indices()
        shard = self.shards()
        tasks = self.tasks()
        reds = set([detail['index'] for detail in indices if detail['health'] == 'red'])
        if reds:
            self.alarms.add(message.索引red.format(", ".join(reds)))

        yellows = set([detail['index'] for detail in indices if detail['health'] == 'yellow'])
        if yellows:
            self.alarms.add(message.索引yellow.format(", ".join(yellows)))

        longtasks = [detail for detail in tasks if detail['running_time'].endswith('h')]
        if longtasks:
            for t in longtasks:
                self.alarms.add(message.ES长任务.format(t['running_time'], t['action']))

        for item in indices:
            index = self.is_read_only(item['index'])
            if index and index.strip() != '':
                self.read_only_indices.append(index)

        if len(self.read_only_indices) > 0:
            self.alarms.add(message.ES索引只读.format(",".join(self.read_only_indices)))
