from concurrent.futures import ThreadPoolExecutor
import config
import message
from checkers.cluster import Cluster
from core.checker import BaseChecker, CheckerManager, share
from core.textparser import TextParser


class Kafka(BaseChecker):

    def __init__(self):
        super().__init__("Kafka")

    def run(self) -> None:
        props = CheckerManager.share(Cluster, 'db_bigdata')
        slaves = props['slaves']
        zookeepers = ','.join(slaves)
        brokers = ','.join([s + ':9092' for s in slaves])
        topics = self.topics(zookeepers)
        groups = self.groups(brokers)
        lags = self.lags(brokers, groups)
        for group, topics in lags.items():
            for topic, detail in topics.items():
                max_lag = detail['max_lag']
                min_lag = detail['min_lag']
                total_lag = detail['total_lag']
                if max_lag > config.kafka_max_partition_lag:
                    self.alarms.add(message.Kafka数据堆积.format(group, topic, min_lag, max_lag, total_lag))

    @share
    def lags(self, brokers, groups) -> dict:
        """
        只在主节点上执行该方法
        """
        if self.should_run():
            self.initial_info_collection_file()
            rs = {}
            pool = ThreadPoolExecutor(max_workers=10)
            for group in groups:
                pool.submit(self.group_offset, brokers, group, rs)
            pool.shutdown()

            return rs

        return {}

    @share
    def group_offset(self, brokers, group, rs):
        out = self.cmd(
            f"/bigdata/salut/components/kafka/bin/kafka-consumer-groups.sh --bootstrap-server {brokers} --describe --group {group}")
        if out.code == 0:
            out = TextParser(
                out.stdout,
                strip_header='TOPIC{:s}PARTITION{:s}CURRENT-OFFSET{:s}LOG-END-OFFSET{:s}LAG{:s}CONSUMER-ID{:s}HOST{:s}CLIENT-ID')
            lines = out.parse_rows(
                '{topic}{:s}{partition}{:s}{current}{:s}{end}{:s}{lag:d}{:s}{consumer}{:s}{host}{:s}{client}',
                allow_missing=True)
            for l in lines:
                topic = l['topic']
                lag = l['lag']
                current = rs.setdefault(group, {
                    topic: {
                        'max_lag': lag,
                        'min_lag': lag,
                        'total_lag': 0
                    }
                })
                current[topic] = {
                    'max_lag': max(lag, current[topic]['max_lag']),
                    'min_lag': min(lag, current[topic]['min_lag']),
                    'total_lag': lag + current[topic]['total_lag']
                }

    @share
    def topics(self, zookeepers) -> list:
        """
        只在主节点上执行该方法
        """
        if self.should_run():
            self.initial_info_collection_file()
            out = self.cmd(
                f"/bigdata/salut/components/kafka/bin/kafka-topics.sh --zookeeper {zookeepers} --list")
            if out.code == 0:
                current_topics = [topic.strip() for topic in out.stdout.split('\n') if topic.strip() != '']
                if len(current_topics) < len(config.kafka_all_topics):
                    lost_topics = set(config.kafka_all_topics) - set(current_topics)
                    self.alarms.add(message.Kafka当前缺失的Topic.format(', '.join(lost_topics)))

                return current_topics

        return []

    @share
    def groups(self, brokers) -> list:
        """
        只在主节点上执行该方法
        """
        if self.should_run():
            self.initial_info_collection_file()
            out = self.cmd(
                f"/bigdata/salut/components/kafka/bin/kafka-consumer-groups.sh --bootstrap-server {brokers} --list")
            if out.code == 0:
                return [group.strip() for group in out.stdout.split('\n') if group.strip() != '']

        return []
