import json
import textwrap
import time
from abc import ABCMeta, abstractmethod
from threading import Lock
import config
from core import common
from core.common import CmdContext, HttpContext, log, FileReadContext, CaseXmlWriter, TextWriter, Msg, AdvanceDict
from core.mock import mock_cmd, mock_http, mock_file
from core.reflection import get_class_by_full_name
from core.cluster import ClusterType


def share(func):
    """
    Checker中的方法添加此装饰器后，被调用后的返回值会被自动保存在该Checker的share成员变量中。

    share成员变量作用：

        1 share成员变量中的内容会被作为关键信息输出到文件中。

        2 使用CheckerManager.share函数，可以获取到其他Checker的share变量

    :param func:
    :return:
    """

    def wrapper(*args, **kwargs):
        if len(args) == 0 or not isinstance(args[0], BaseChecker):
            raise BaseException('装饰器仅适用于BaseChecker子类的实例方法')
        checker: BaseChecker = args[0]
        value = func(*args, **kwargs)
        checker.share[func.__name__] = value
        return value

    return wrapper


class Alarms:
    def __init__(self):
        self._messages = {}

    @property
    def message(self):
        return self._messages

    def add(self, msg: Msg):
        # 如果msg.leve不在self._messages中将字典的值设为空列表
        self._messages.setdefault(msg.level, []).append(msg)

    def pretty(self):
        levels = []
        for level, msgs in self._messages.items():
            fmt_message = "\n".join(
                ['{:>3d}. {}'.format(i + 1, m.description) for i, m in enumerate(msgs)])
            levels.append(textwrap.dedent(f'''
            ==============
              {level}
            ==============
            ''') + fmt_message)
        return '\n'.join(levels)

    def __str__(self):
        return self.pretty()

    def __repr__(self):
        return self.pretty()


class BaseChecker(metaclass=ABCMeta):

    def __init__(self, title):
        self.title = title
        self._lock = Lock()
        self.alarms = Alarms()
        self.share = AdvanceDict()
        self.enable = True
        self.has_check = False
        basedir = None
        self.operation_writer = None
        self.share_writer = None
        self.alarm_writer = None

    @abstractmethod
    def run(self) -> None:
        """
        健康状态检测，此处不应包括信息收集逻辑
        """
        pass

    def initial_info_collection_file(self):
        """
        初始化信息收集结果、故障告警及日志记录的文件
        """
        basedir = f'{common.project_root}/{config.tmp_output_dir}'
        self.operation_writer = CaseXmlWriter(f'{basedir}/信息收集/{self.title}.xml')
        self.share_writer = TextWriter(f'{basedir}/信息收集/{self.title}.json')
        self.alarm_writer = TextWriter(f'{basedir}/故障告警/{self.title}.txt')

    def should_run(self) -> bool:
        # 通过反射获取
        cluster_class = get_class_by_full_name(config.cluster_type)
        cluster: ClusterType = cluster_class()
        cluster_info = cluster.get_cluster_ips()
        local_ip = cluster_info.get('local_ip')
        master_ip = cluster_info.get('master')
        return local_ip == master_ip

    def check(self) -> None:
        self._lock.acquire()
        try:
            if not self.has_check:
                s = time.time()
                self.run()
                if self.share_writer is not None:
                    self.share_writer.write(json.dumps(self.share, ensure_ascii=False, indent=2, default=str))
                if self.alarms.message:
                    self.alarm_writer.write(self.alarms.pretty())
                log.info('诊断完成，耗时：{:.0f} s {} {}'.format(time.time() - s, self.__module__, self.title))
        finally:
            self.has_check = True
            self._lock.release()
            if self.operation_writer is not None:
                self.operation_writer.close()
            if self.share_writer is not None:
                self.share_writer.close()
            if self.alarm_writer is not None:
                self.alarm_writer.close()

    @mock_file('{1}')
    def read_file(self, file) -> FileReadContext:
        out = common.read_file(file)
        self.operation_writer.add(out.to_dom())
        return out

    @mock_cmd('{1}')
    def cmd(self, cmd) -> CmdContext:
        out = common.execute(cmd)
        self.operation_writer.add(out.to_dom())
        return out

    @mock_cmd('{3}')
    def gpsql(self, database, passwd, sql: str, user='gpadmin', port=6543) -> CmdContext:
        return self.pgsql(database, passwd, sql, user, port)

    @mock_cmd('{3}')
    def pgsql(self, database, passwd, sql: str, user='postgres', port=5432) -> CmdContext:
        return self.cmd(f"""
            su - {user} -c "export PGPASSWORD={passwd};psql -U {user} -p {port} -d {database} -c \\\"{sql}\\\""
        """)

    @mock_cmd('ssh {1} {2}')
    def ssh(self, ip, cmd) -> CmdContext:
        return common.ssh(ip, cmd)

    @mock_http('{1} {4}')
    def http(self, method, ip, port, path, payload='', headers=None) -> HttpContext:
        output = common.http_request(method, ip, port, path, payload, headers)
        self.operation_writer.add(output.to_dom())
        return output


class CheckerManager:
    """
    保存已经校验完成的校验类实例
    dict(类->对象)
    """
    _instances = {}


    @staticmethod
    def get_or_create_checker(clazz) -> BaseChecker:
        checker = CheckerManager._instances.get(clazz)
        if checker is None:
            checker = clazz()
            CheckerManager._instances[clazz] = checker
        return checker


    @staticmethod
    def share(clazz, *key) -> list[dict]:
        obj: BaseChecker = CheckerManager.get_or_create_checker(clazz)
        obj.check()  # 会调用类中的run方法
        if len(key) == 0:
            return obj.share
        else:
            return obj.share.gets(*key)
