import datetime
import http.client
import logging
import os
import socket
import subprocess
import traceback
from dataclasses import dataclass
from threading import Lock
from typing import List
from xml.dom import minidom
import re
import config


lock = Lock()
pattern = r"\x1B\[([0-9]{1,2}(;[0-9]{1,2})?)?[m|K]"


def now():
    return datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')


def get_logger(file):
    # 创建tmp目录
    os.makedirs(os.path.dirname(file), exist_ok=True)
    log = logging.getLogger()
    log.setLevel(logging.INFO)
    fmt = logging.Formatter('[%(asctime)s][%(levelname)s][%(module)s.%(funcName)s()] %(message)s',
                            datefmt='%Y-%m-%d %H:%M:%S')
    file_handler = logging.FileHandler(file)
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(fmt)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(fmt)
    log.addHandler(file_handler)
    log.addHandler(stream_handler)
    return log


project_root = os.path.dirname(__file__) + '/../'
log = get_logger(f'{project_root}/{config.tmp_output_dir}/{config.log_file_name}')


@dataclass
class FileReadContext:
    path: str
    content: str
    time = now()


    def to_dom(self):
        dom = minidom.Document()
        case_node = dom.createElement(config.xml_tag_operation)
        case_node.setAttribute('id', self.path)
        case_node.setAttribute('type', 'file')
        case_node.setAttribute('time', self.time)

        output_node = dom.createElement(config.xml_tag_output)
        output_node.appendChild(dom.createTextNode(self.content))
        case_node.appendChild(output_node)
        return case_node


@dataclass
class CmdContext:
    cmd: str
    stdout: str
    stderr: str
    code: int
    time = now()


    @property
    def success(self):
        return self.code == 0


    def to_dom(self):
        dom = minidom.Document()
        case_node = dom.createElement(config.xml_tag_operation)
        case_node.setAttribute('id', self.cmd)
        case_node.setAttribute('type', 'cmd')
        case_node.setAttribute('time', self.time)

        output_node = dom.createElement(config.xml_tag_output)
        output_node.appendChild(dom.createTextNode(self.stdout))
        case_node.appendChild(output_node)

        code_node = dom.createElement(config.xml_tag_error)
        code_node.appendChild(dom.createTextNode(self.stderr))
        case_node.appendChild(code_node)

        code_node = dom.createElement(config.xml_tag_code)
        code_node.appendChild(dom.createTextNode(str(self.code)))
        case_node.appendChild(code_node)
        return case_node


    def __repr__(self):
        return f'''
[{now()}]
{self.cmd}
=====================================================
{self.stdout}
-----------------------------------------------------
标准错误输出
-----------------------------------------------------
{self.stderr}
-----------------------------------------------------
退出码：{self.code}
-----------------------------------------------------
'''

    def __str__(self):
        return self.__repr__()


@dataclass
class HttpContext:
    header: str
    body: str
    url: str
    output: str
    code: int
    time = now()


    @property
    def success(self):
        return self.code == 200


    def to_dom(self):
        dom = minidom.Document()
        case_node = dom.createElement(config.xml_tag_operation)
        case_node.setAttribute('id', self.url)
        case_node.setAttribute('type', 'http')
        case_node.setAttribute('time', self.time)

        output_node = dom.createElement(config.xml_tag_http_body)
        output_node.appendChild(dom.createTextNode(self.output))
        case_node.appendChild(output_node)

        output_node = dom.createElement(config.xml_tag_output)
        output_node.appendChild(dom.createTextNode(self.output))
        case_node.appendChild(output_node)

        code_node = dom.createElement(config.xml_tag_code)
        code_node.appendChild(dom.createTextNode(str(self.code)))
        case_node.appendChild(code_node)
        return case_node


    def __repr__(self):
        return f'''
{self.url}
{self.header}
{self.body}
=====================================================
{self.output}
-----------------------------------------------------
状态码：{self.code}
-----------------------------------------------------'''

    def __str__(self):
        return self.__repr__()


def del_empty(text: str):
    return '\n'.join([line.strip() for line in text.split('\n') if line.strip()])


def read_file(file: str) -> FileReadContext:
    with open(file, encoding='utf-8') as f:
        output = f.read()
        return FileReadContext(file, output)


def execute(cmd, raise_exception=False) -> CmdContext:
    try:
        with subprocess.Popen(cmd, stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE, shell=True) as p:
            stdout, stderr = p.communicate()
            # 通过正则表达式将shell命令输出结果中的色彩字符过滤掉
            stdout_filter = re.sub(pattern, '', stdout.decode('utf-8'))
            stderr_filter = re.sub(pattern, '', stderr.decode('utf-8'))
            ret = CmdContext(cmd, stdout_filter, stderr_filter, p.returncode)
            if raise_exception and not ret.success:
                raise BaseException(f'{cmd}命令执行失败\n{ret}')
            return ret
    except:
        log.error(traceback.format_exc())


def ssh(ip: str, cmd: str, raise_exception=False):
    cmd = cmd.replace('"', r'\"')
    return execute(f'ssh {ip} "{cmd}"', raise_exception)


def http_request(method, ip, port, path, payload='', headers=None) -> HttpContext:
    if headers is None:
        headers = {}
    conn = http.client.HTTPConnection(ip, port)
    conn.request(method, url=path, body=payload, headers=headers)
    res = conn.getresponse()
    data = res.read()
    return HttpContext(f'{method} http://{ip}:{port}{path}', headers, payload, data.decode("utf-8"), res.getcode())


def get_local_ips() -> List[str]:
    return socket.gethostbyname_ex(socket.gethostname())[2]


class CaseXmlWriter:
    def __init__(self, path):
        self.dom = minidom.Document()
        self.root = self.dom.createElement('Cases')
        self.dom.appendChild(self.root)
        self.path = path
        dirname = os.path.dirname(self.path)
        if not os.path.exists(dirname):
            os.makedirs(dirname, exist_ok=True)
        if os.path.exists(self.path):
            os.remove(self.path)

    def add(self, child):
        self.root.appendChild(child)

    def close(self):
        with open(self.path, 'w', encoding='utf-8') as f:
            self.dom.writexml(f, indent='', addindent='    ', newl='\n', encoding='UTF-8')


class TextWriter:

    def __init__(self, path):
        self.path = path
        self.file = None
        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname, exist_ok=True)
        if os.path.exists(path):
            os.remove(path)

    def write(self, text):
        self.file = self.file if self.file is not None else open(self.path, 'w', encoding='utf-8')
        self.file.write(text)
        self.file.flush()

    def write_line(self, text):
        self.write(text + '\n')

    def close(self):
        if self.file is not None:
            self.file.close()


@dataclass
class Msg:
    level: str
    description: str

    def format(self, *args):
        return Msg(self.level, self.description.format(*args))


class AdvanceDict(dict):
    """
        功能1：取出key=db_bigdata的内层字典
    """
    def get(self, key) -> dict:
        value = self

        return value[key]

    def gets(self, *keys) -> list[dict]:
        if len(keys) == 1:
            return self.get(keys[0])
        return list([self.get(key) for key in keys])
