import json
import threading
from http.server import HTTPServer, BaseHTTPRequestHandler
import config
from core import common
from core.checker import CheckerManager
from core.common import log
from core.reflection import get_class_by_full_name


class HttpHandler(BaseHTTPRequestHandler):
    timeout = 5

    def do_GET(self):
        self.handle_request('get')

    def do_POST(self):
        data = self.rfile.read(int(self.headers['content-length']))  # 固定格式，获取表单提交的数据
        body = data.decode('utf-8')
        body_dict = json.loads(body)
        self.handle_request('post', body_dict)

    def handle_request(self, method, body_dict=None):
        match method, self.path:
            case ('post', '/share/report'):
                result = self.share_report(body_dict)
                self.send_response(200)
            case ('post', '/share/fetch'):
                result = self.share_fetch(body_dict)
                self.send_response(200)
            case _:
                result = '页面未找到'
                self.send_response(404)
        self.send_header("Content-type", "text/json")
        self.end_headers()
        self.wfile.write(result.encode())

    def share_report(self, body: dict):
        match body:
            case {'checker': checker, 'share': share}:
                checker_class = get_class_by_full_name(checker)
                checker = CheckerManager.get_or_create_checker(checker_class)
                checker.share = share
                log.info(f'收到来自 {self.request.host} {checker} 的共享信息')
                return 'ok'
            case _:
                return f'入参格式不合法：{body}'

    def share_fetch(self, body: dict):
        match body:
            case {'checker': checker}:
                checker_class = get_class_by_full_name(checker)
                share = CheckerManager.share(checker_class)
                log.info(f'{self.request.host} 请求获取 {checker} 的共享信息')
                return json.dumps(share)
            case _:
                return f'入参格式不合法：\n{body}'


class HttpServer:

    def __init__(self, port: int, bind='0.0.0.0'):
        self.host = (bind, port)
        self._server = HTTPServer(self.host, HttpHandler)

    def serve(self):
        print("启动HTTP服务端: %s:%s" % self.host)
        self._server.serve_forever()

    def shutdown(self):
        self._server.shutdown()

    def start(self):
        threading.Thread(target=self.serve, name='http-server-main').start()


def report_to_master(checker, master_ip):
    class_name = checker.__class__.__name__
    module_name = checker.__module__
    out = common.http_request('POST', master_ip, config.http_server_port, '/share/report', payload=json.dumps({
        'checker': f'{class_name}.{module_name}',
        'share': checker.share
    }))
    if not out.success:
        log.error(f'上报 {class_name} 共享信息失败, {out.output}')


def fetch_from_master(checker, master_ip):
    class_name = checker.__class__.__name__
    module_name = checker.__module__
    log.info(f'{class_name}不支持在当前节点运行，尝试从主机获取共享信息')
    out = common.http_request('POST', master_ip, config.http_server_port, '/share/fetch', payload=json.dumps({
        'checker': f'{class_name}.{module_name}'
    }))
    if out.success:
        checker.share = out.output
    else:
        log.error(f'获取 {class_name} 共享信息失败, {out.output}')


if __name__ == '__main__':
    config.mock_enable = True
    HttpServer(11111).serve()
