import os.path
import xml.dom.minidom
from pyexpat import ExpatError
import config
from core import parse
from core.common import CmdContext, HttpContext, FileReadContext


def mock_cmd(key_pattern: str):
    return mock(key_pattern, 'cmd')


def mock_http(key_pattern: str):
    return mock(key_pattern, 'http')


def mock_file(key_pattern: str):
    return mock(key_pattern, 'file')


def mock(key_pattern: str, mock_type):
    """
    模拟数据装饰器，用于根据函数入参生成唯一标识case_id，根据case_id从配置文件中获取模拟输出

    举例：
        被装饰函数定义为：http_request(method, ip, port, url)

        实际入参为："GET", "127.0.0.1", "9200", "/_cat/indices"

        key_pattern = "{0} {3}"

        则生成的case_id为："GET /_cat/indices"


    :param key_pattern: case_id表达式
    :param mock_type: 模拟类型，支持cmd（命令行）和rest（接口调用）
    :return:
    """

    def generate(func):
        def wrapper(*args, **kwargs):
            caller = args[0]
            if config.mock_enable:
                case_id = key_pattern.format(*args)
                return do_mock(caller, case_id, mock_type)
            else:
                return func(*args, **kwargs)

        return wrapper

    return generate


def do_mock(caller, case_id, mock_type):
    file = mock_file_path(caller)
    try:
        tree: xml.dom.minidom.Document = xml.dom.minidom.parse(file)
    except ExpatError as e:
        raise BaseException(f'mock文件{file}解析异常', e)
    case = find_case(tree, case_id)
    if case is None:
        raise BaseException(f'mock文件{file}, 中无法找到id为 {case_id} 的用例')
    output = xml_tag_text(case, config.xml_tag_output)
    match mock_type:
        case 'cmd':
            error = xml_tag_text(case, config.xml_tag_error, '')
            code = xml_tag_text(case, config.xml_tag_code, 0)
            code = int(code)
            return CmdContext(case_id, output, error, code)
        case 'http':
            code = xml_tag_text(case, config.xml_tag_code, 200)
            code = int(code)
            return HttpContext(case_id, '', '', output, code)
        case 'file':
            return FileReadContext(case_id, output)
        case _:
            raise BaseException(f'未找到唯一标识为 {case_id} 的用例')


def mock_file_path(checker):
    return os.path.dirname(__file__) + '/../' + config.mock_dir + '/' + checker.__module__.split(".")[-1] + '.xml'


def xml_tag_text(parent, tag, default=None):
    elem = parent.getElementsByTagName(tag)
    if not elem:
        if default is not None:
            return default
        else:
            raise BaseException(f'未找到XML标签{tag}')
    tag = elem[0]
    if len(tag.childNodes) > 0:
        text = tag.childNodes[0].data
        return text


def find_case(tree: xml.dom.minidom.Document, case_id):
    case = tree.getElementById(case_id)

    if case:
        return case

    cases = list([case for case in tree.getElementsByTagName('Case') if
                  parse.parse(case.getAttribute('id'), case_id)])
    if len(cases) > 0:
        return cases[0]
