from core import parse


class TextParser:

    def __init__(self, target_text: str,
                 strip_line=True,
                 strip_header=None,
                 strip_tail=None):
        self.target_text = target_text
        self.strip_each_line = strip_line
        self.strip_header = strip_header
        self.strip_tail = strip_tail
        self.extra_types = dict(
            comma_split=comma_split,
            nullable=nullable_string
        )
        self.striped_lines = self._striped_lines()
        self.text = '\n'.join(self._striped_lines())

    def _striped_lines(self):
        lines = []
        start = False
        for line in self.target_text.split('\n'):
            if self.strip_each_line:
                line = line.strip()
            if line.strip() == '\n':
                # 跳过头部空行
                continue
            if not self.strip_header:
                start = True
            if self.strip_header and parse.search(self.strip_header, line):
                # 跳过表头
                start = True
                continue
            if self.strip_tail and parse.search(self.strip_tail, line):
                break
            if start:
                lines.append(line.strip())

        return lines

    def parse(self, *patterns, full_match=True):
        """
        多个表达式各自匹配一次文本，匹配到的结果以键值对放进字典，每个表达式仅会命中一次。

        注意：
            1. 在编写表达式时，多个表达式不要使用同一个key，否则会造成结果覆盖。
            2. 每行部分符合表达式即可匹配成功

        :param full_match: 是否整行匹配，如果为False，则一行中只要子串匹配就可以
        :param patterns: 待匹配表达式，可多个
        :return: dict(key->value)
        """
        result = {}
        funcs = list([self._parse_func(full_match, p) for p in patterns])
        for line in self.striped_lines:
            for func in funcs:
                r = func(line)
                if r and r.named:
                    result.update(r.named)
        return result

    def parse_rows_by_spilt(self, separator, names, strip=True) -> list:
        rs = []
        for line in self.striped_lines:
            d = {}
            values = line.split(separator)
            if len(values) != len(names):
                raise BaseException(f'预期字段数量与实际分隔后的值的数量不符，字段: {names}, 值: {values}')
            for i, name in enumerate(names):
                raw_val = values[i]
                val = raw_val.strip() if strip else raw_val
                d[name] = val
            rs.append(d)
        return rs

    def parse_rows(self, pattern, full_match=True, allow_missing=False) -> list:
        """
        用表达式匹配文本中的每行，匹配到的结果以键值对放进列表。

        :param allow_missing: 是否运行无匹配结果
        :param full_match: 是否整行匹配，如果为False，则一行中只要子串匹配就可以
        :param pattern: 匹配表达式
        :return: list(dict(key->value))
        """
        func = self._parse_func(full_match, pattern)
        result = []
        for line in self.striped_lines:
            r = func(line)
            if r and r.named:
                result.append(r.named)
        if not allow_missing and len(result) == 0:
            self._match_failed(pattern)
        return result

    def _parse_func(self, full_match, pattern):
        p = parse.compile(pattern, self.extra_types)
        if full_match:
            func = p.parse
        else:
            func = p.search
        return func

    def _match_failed(self, p):
        raise BaseException(f'匹配失败,表达式为：\n{p}\n待匹配文本为：\n{self.text}')


    def parse_json(self, dicts:dict) -> list:
        """
        解析多层嵌套的字典，取出内层的列表
        """
        result = None
        for key, value in dicts.items():
            num = len(value)
            if num > 1:
                result = value
                break
            result = self.parse_json(value)
        return result


def comma_split(string: str):
    return list([seg.strip() for seg in string.strip().split(',')])


def nullable_string(string: str):
    return string.strip()


nullable_string.pattern = r'\s*(^|)+\s*'
