import argparse
import datetime
import os
import textwrap
import time
import traceback
from concurrent.futures import ThreadPoolExecutor

import config
from core import common
from core.checker import BaseChecker, CheckerManager
from core.cluster import ClusterType
from core.common import ssh, execute, log
from core.httpserver import HttpServer
from core.reflection import scan_class, get_class_by_full_name, get_cluster_info

start_timestamp = time.time()
# 结果临时输出目录
path_remote_tmp_output_dir = f'{config.dispatch_path}/{config.tmp_output_dir}'
path_local_tmp_output_dir = f'{common.project_root}/{config.tmp_output_dir}'
# 结果输出目录
path_remote_output_dir = f'{config.dispatch_path}/{config.output_dir}'
path_local_output_dir = f'{common.project_root}/{config.output_dir}'
# 通过反射获取类
cluster_class = get_class_by_full_name(config.cluster_type)
# 实例化
cluster: ClusterType = cluster_class()
# 在core.common.get_loggerz中会创建输出日志的目录
log.info(f'开始诊断集群: {cluster.get_name()}')


def output_dir(time: datetime):
    path_output_root = path_local_output_dir
    return f'{path_output_root}/{time.strftime("%y-%m-%d-%H:%M:%S")}'


def run_local_node(args):
    """
        开启config.max_thread_num个线程查询各个模块里的数据
    """
    try:
        pool = ThreadPoolExecutor(max_workers=config.max_thread_num)
        for clazz in scan_class(cluster.get_checkers_package(), BaseChecker):
            checker: BaseChecker = CheckerManager.get_or_create_checker(clazz)
            pool.submit(check, checker)
        pool.shutdown()
    except:
        log.error(traceback.format_exc())


def run_remote_node(args, ip, start_datetime):
    try:
        ssh(ip, f'rm -rf {config.dispatch_path} && mkdir -p {config.dispatch_path}', raise_exception=True)
        start = time.time()
        execute(
            f'rsync -a --exclude={config.output_dir} --exclude={config.tmp_output_dir} {common.project_root}/* {ip}:{config.dispatch_path}',
            raise_exception=True)
        log.info(f'{ip}: 文件拷贝完成，耗时：{int(time.time() - start)} s')
        start = time.time()
        ssh(ip,
            f'cd {config.dispatch_path} && bin/mini-python3.10.6/bin/python3 main.py --local',
            raise_exception=True)
        log.info(f'{ip}：诊断完成，耗时：{int(time.time() - start)} s')
        start = time.time()
        execute(f'rsync -a {ip}:{path_remote_tmp_output_dir}/* {output_dir(start_datetime)}/{ip}',
                raise_exception=True)
        log.info(f'{ip}：日志拷贝完成，耗时：{int(time.time() - start)} s')
        # ssh(ip, f'rm -rf {config.dispatch_path} ', raise_exception=True)
    except:
        log.error(traceback.format_exc())


def run(args):
    if args.mock:
        config.mock_enable = True
        run_local = True
        log.info('当前使用模拟数据运行，使用本地模式')
    else:
        run_local = args.run_local
    cluster_info = get_cluster_info()
    nodes = list(item['ip'] for item in cluster_info.get('nodes'))
    local_ip = cluster_info.get('local_ip')
    master_ip = cluster_info.get('master')
    start_datetime = datetime.datetime.now()
    if run_local:
        run_local_node(args)
        if master_ip == local_ip:
            os.makedirs(output_dir(start_datetime), exist_ok=True)
            # 移动本节点的输出到集群输出路径
            execute(f'mv {path_local_tmp_output_dir} {output_dir(start_datetime)}/{local_ip}-master',
                    raise_exception=True)
    else:
        run_cluster(args, nodes, master_ip, start_datetime)

    run_over(config.output_dir)


def run_cluster(args, nodes, master_ip, start_datetime):
    server = HttpServer(port=config.http_server_port)
    server.start()
    try:
        # # 各个节点会并发往此目录发送信息收集结果，必须提前把目录建好
        os.makedirs(output_dir(start_datetime), exist_ok=True)
        pool = ThreadPoolExecutor(config.max_thread_num)
        # 主节点的信息单独收集
        pool.submit(run_local_node, args)
        for ip in nodes:
            # 跳过当前主节点
            if ip == master_ip:
                continue
            # 分发可执行文件，并在集群节点上执行信息收集
            pool.submit(run_remote_node, args, ip, start_datetime)
        # 执行本节点的信息收集，当前节点无需远程拷贝可执行文件，因此单独执行
        pool.shutdown()
        # 移动本节点的输出到集群输出路径
        execute(f'mv {path_local_tmp_output_dir} {output_dir(start_datetime)}/{master_ip}-master',
                raise_exception=True)
    finally:
        server.shutdown()


def run_over(out_dir: str):
    log.info(textwrap.dedent(f'''
 ===================================================================================
 诊断完成，耗时 {int(time.time() - start_timestamp)} s， 请从 {out_dir} 目录获取诊断结果
 ===================================================================================
'''))


def check(checker: BaseChecker):
    try:
        # if checker.should_run():
        checker.check()
    except:
        log.error(traceback.format_exc())


def main():
    args = pars_args()
    run(args)


def pars_args():
    parser = argparse.ArgumentParser(description='集群信息收集及诊断工具')
    parser.add_argument('--mock', dest='mock', action='store_true',
                        help='是否使用模拟数据运行，使用模拟数据时默认启用--local选项')
    parser.add_argument('--local', dest='run_local', action='store_true',
                        help='本地模式，仅在当前节点执行，默认为在整个集群执行')
    parser.add_argument('--module', dest='target_module',
                        help='指定需要执行的模块')
    parser.add_argument('--class', dest='target_class',
                        help='指定需要执行的类')
    parser.add_argument('--square', dest='square', type=int,
                        help='指定需要执行的类')
    args = parser.parse_args()

    return args


if __name__ == '__main__':
    main()
